package combinatorParser.comprehension;

object Parsers {

  import Tokens._;
  
  abstract class Parser[T] {
    type Input = List[Token];
    type Result = Option[Pair[T, Input]];
    def apply(in: Input): Result;
    def filter(pred: T => boolean) = new Parser[T] {
      def apply(in: Input): Result = Parser.this.apply(in) match {
        case None => None
        case Some(Pair(x, in1)) =>
          if (pred(x)) Some(Pair(x, in1)) else None
      }
    }
    def map[S](f: T => S) = new Parser[S] {
      def apply(in: Input): Result = Parser.this.apply(in) match {
        case None => None
        case Some(Pair(x, in1)) => Some(Pair(f(x), in1))
      }
    }
    def flatMap[S](f: T => Parser[S]) = new Parser[S] {
      def apply(in: Input): Result = Parser.this.apply(in) match {
        case None => None
        case Some(Pair(x, in1)) => f(x).apply(in1)
      }
    }
    def & [S](q: => Parser[S]): Parser[S] = 
      for (val _ <- this; val x <- q) yield x;
    def | (q: => Parser[T]) = new Parser[T] {
      def apply(in: Input): Result = Parser.this(in) match {
        case r @ Some(_) => r
        case None => q(in)
      }
    }
  }
  
  def rep[T](p: Parser[T]): Parser[List[T]] =
    rep1(p) | success(List());
  
  def rep1[T](p: Parser[T]): Parser[List[T]] =
    for (val x <- p; val xs <- rep(p)) yield x :: xs;
  
  def opt[T](p: Parser[T]): Parser[List[T]] =
    (for (val x <- p) yield
      List(x)) | success(List());

  def token(tclass: int) = new Parser[String] {
    def apply(in: Input): Result = 
      if (!in.isEmpty && in.head.tclass == tclass) 
        Some(Pair(in.head.chars, in.tail))
      else None
  }

  def failure = new Parser[All] {
    def apply(in: Input): Result = None
  }
  
  def success[T](x: T) = new Parser[T] {
    def apply(in: Input): Result = Some(Pair(x, in))
  }
    
}
  
