package combinatorParser.generic;

abstract class Parsers {

  import Tokens._;
  
  type Input = List[Token];
  type Result = Option[Pair[List[Any], Input]];

  def reduce(elems: List[Any]): List[Any];

  abstract class Parser {
    def apply(in: Input): Result;
    def & (q: => Parser) = new Parser {
      def apply(in: Input): Result = Parser.this(in) match {
        case Some(Pair(x, in1)) => q(in1) match {
          case Some(Pair(y, in2)) => Some(Pair(x ::: y, in2))
          case None => None
        }
        case None => None
      }
    }
    def | (q: => Parser) = new Parser {
      def apply(in: Input): Result = Parser.this(in) match {
        case r @ Some(_) => r
        case None => q(in)
      }
    }
    def !! (f: List[Any] => List[Any]) = new Parser {
      def apply(in: Input): Result = Parser.this(in) match {
        case r @ Some(Pair(x, in)) => Some(Pair(f(x), in))
        case None => None
      }
    }
    def ! = !!(reduce)
  }

  def token(tclass: int) = new Parser {
    def apply(in: Input): Result =  
      if (!in.isEmpty && in.head.tclass == tclass) 
        Some(Pair(List(in.head.chars), in.tail))
      else None
  }

  def failure = new Parser {
    def apply(in: Input): Result = None
  }
  
  def success(x: List[Any]) = new Parser {
    def apply(in: Input): Result = Some(Pair(x, in))
  }  

  def empty = success(List());

  def rep(p: Parser): Parser = p & rep(p) | empty;
  def rep1(p: Parser): Parser = p & rep(p);
  def opt(p: Parser): Parser = p | empty;

}
  
