package combinatorParser.uniform;

object Parsers {

  import Tokens._;

  type Input = List[Token];
  type Result = Option[Pair[Any, Input]];

  abstract class Parser {
    def apply(in: Input): Result;
    def & (q: => Parser) = new Parser {
      def apply(in: Input): Result = Parser.this(in) match {
        case Some(Pair(x, in1)) => q(in1) match {
          case Some(Pair(xs, in2)) => Some(Pair(compose(x, xs), in2))
          case None => None
        }
        case None => None
      }
    }
    def | (q: => Parser) = new Parser {
      def apply(in: Input): Result = Parser.this(in) match {
        case r @ Some(_) => r
        case None => q(in)
      }
    }
  }

  def token(tclass: int) = new Parser {
    def apply(in: Input): Result = 
      if (!in.isEmpty && in.head.tclass == tclass) 
        Some(Pair(in.head.chars, in.tail))
      else None
  }

  def rep(p: Parser) = new Parser {
    def apply(in: Input): Result = p(in) match {
      case Some(Pair(x, in1)) => 
        apply(in1) match {
          case Some(Pair(xs, in2)) => 
            Some(Pair(compose(x, xs), in2))
          case None => Some(Pair(List(x), in1))
        }
      case None => Some(Pair(null, in))
    }
  }

  def opt(p: Parser) = new Parser {
    def apply(in: Input): Result = p(in) match {
      case r @ Some(Pair(x, in1)) => r
      case None => Some(Pair(null, in))
    }
  }

  def failure = new Parser {
    def apply(in: Input): Result = None
  }
  
  def success(x: Any) = new Parser {
    def apply(in: Input): Result = Some(Pair(x, in))
  }  

  def empty = success(List());

  def compose(hd: Any, tl: Any) = 
    if (hd == null) tl
    else if (tl == null) hd 
    else Pair(hd, tl)
}
  
